from __future__ import annotations

import ctypes
import collections.abc as abc
import math

from typing import (
    List,
    ClassVar,
    Type,
    Generic,
    TypeVar,
    Union,
    Iterable,
    Optional,
    overload,
    Type,
    Sequence,
    Iterator,
)
from abc import abstractmethod, ABCMeta
from contextlib import contextmanager
from contextvars import ContextVar


from .message_base import MessageBase

__all__ = [
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Byte",
    "Uint8",
    "Uint16",
    "Uint32",
    "Uint64",
    "Float",
    "Double",
    "Char",
    "Struct",
    "String",
    "ByteArray",
    "IntArray",
    "FloatArray",
    "StructArray",
    "disable_message_validation",
]

_VALIDATION_ENABLED: ContextVar[bool] = ContextVar("_VALIDATION_ENABLED", default=True)


@contextmanager
def disable_message_validation(ignore=False):
    """Context manager function to temporarily disable message field validation
    Use with `with` keyword:
    `with disable_message_validation():`

    Optionally pass in ignore=True to do nothing, e.g. for debugging:

    ```
    DEBUG = True
    with disable_message_validation(ignore=DEBUG):
        ... # disable validation unless DEBUG is True
    ```
    """
    if not ignore:
        token = _VALIDATION_ENABLED.set(False)
        yield
        _VALIDATION_ENABLED.reset(token)
    else:
        yield  # dummy context


_P = TypeVar("_P")  # Parent
_V = TypeVar("_V")  # Value
_C = TypeVar("_C")  # CType


class FieldValidator(Generic[_P, _V], metaclass=ABCMeta):
    """Abstract base class for all message field validator descriptors"""

    def __init__(self) -> None:
        self._ctype: Type[ctypes._CData] = ctypes._CData

    def __set_name__(self, owner: _P, name: str):
        self._owner = owner
        self._public_name = name
        self._private_name = "_" + name

    @abstractmethod
    def __get__(self, obj: _P, objtype=None): ...

    @abstractmethod
    def __set__(self, obj: _P, value: _V): ...

    @abstractmethod
    def validate_one(self, value: _V): ...

    @abstractmethod
    def validate_many(self, value: Iterable[_V]): ...


class FloatValidatorBase(FieldValidator[_P, float], Generic[_P, _C], metaclass=ABCMeta):
    """Abstract base class for float type validators"""

    @abstractmethod
    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_float

    def __get__(self, obj: _P, objtype=None) -> float:
        return getattr(obj, self._private_name)

    def __set__(self, obj: _P, value: Union[float, int, _C]):
        if _VALIDATION_ENABLED.get():
            self.validate_one(value)
        setattr(obj, self._private_name, value)

    def validate_one(self, value: Union[float, int, _C]):
        """Validate a float value

        Args:
            value (float): Float value

        Raises:
            TypeError: Wrong type
            ValueError: Value cannot be precisely represented with this datatype
        """

        if isinstance(value, self._ctype):
            return

        if not isinstance(value, (float, int)):
            raise TypeError(f"Expected {value} to be a float")

        if math.isinf(self._ctype(value).value):
            raise ValueError(
                f"The {value} can not be represented as a {type(self).__name__}"
            )

    def validate_many(self, value: Iterable[Union[float, int]]):
        """Validate multiple float values

        Args:
            value (Iterable[float]): Iterable of floats to validate

        Raises:
            TypeError: Wrong type
            ValueError: Value cannot be precisely represented with this datatype
        """

        # Note: This may not be worth it since this is a rare overflow case.
        try:
            if math.isinf(self._ctype(max(value)).value) or math.isinf(
                self._ctype(min(value)).value
            ):
                raise ValueError(
                    f"{value} contains value(s) that can not be represented as a {type(self).__name__}"
                )
        except TypeError:
            raise TypeError(f"Expected {value!r} to contain all float types.")

    def __repr__(self):
        return f"{type(self).__name__} at 0x{id(self):016X}"


class Float(FloatValidatorBase[_P, ctypes.c_float], Generic[_P]):
    """32-bit Float validator class"""

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_float


class Double(FloatValidatorBase[_P, ctypes.c_double], Generic[_P]):
    """Double (64-bit float) validator class"""

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_double


# Base Class for int validator fields
class IntValidatorBase(FieldValidator[_P, int], Generic[_P, _C], metaclass=ABCMeta):
    """Abstract base class for integer type validators"""

    _size: ClassVar[int] = 1
    _unsigned: ClassVar[bool] = False
    _min: ClassVar[int] = 0
    _max: ClassVar[int] = 2**8 - 1

    @abstractmethod
    def __init__(self, *args): ...

    @property
    def size(self) -> int:
        return self._size

    @property
    def unsigned(self) -> bool:
        return self._unsigned

    @property
    def min(self) -> int:
        return self._min

    @property
    def max(self) -> int:
        return self._max

    def __get__(self, obj: _P, objtype=None) -> int:
        return getattr(obj, self._private_name)

    def __set__(self, obj: _P, value: Union[int, _C]):
        if _VALIDATION_ENABLED.get():
            self.validate_one(value)
        setattr(obj, self._private_name, value)

    def validate_one(self, value: Union[int, _C]):
        """Validate an integer value

        Args:
            value (int): Integer to validate

        Raises:
            TypeError: Wrong type
            ValueError: Integer out of range for this datatype
        """

        if isinstance(value, self._ctype):
            return

        if not isinstance(value, int):
            raise TypeError(f"Expected {value} to be an int")

        if not (self._min <= int(value) <= self._max):
            raise ValueError(
                f"Expected {value} to be in range of {self._min} to {self._max}"
            )

    def validate_many(self, value: Iterable[int]):
        """Validate multiple integer values

        Args:
            value (Iterable[int]): Iterable of integers to validate

        Raises:
            TypeError: Wrong type
            ValueError: Integer out of range for this datatype
        """

        # Check all values
        if any(not isinstance(v, int) for v in value):
            raise TypeError(f"Expected {value} to contain all int types.")

        if (int(max(value)) > self._max) or (min(value) < self._min):
            raise ValueError(
                f"Expected {value} to be in range of {self._min} to {self._max}."
            )

    def __repr__(self):
        return f"{type(self).__name__}(size={self.size}, unsigned={self.unsigned}) at 0x{id(self):016X}"


class Int8(IntValidatorBase[_P, ctypes.c_int8], Generic[_P]):
    """Validator for 8-bit integers"""

    _size: ClassVar[int] = 1
    _unsigned: ClassVar[bool] = False
    _min: ClassVar[int] = -(2**7)
    _max: ClassVar[int] = 2**7 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_int8


class Int16(IntValidatorBase[_P, ctypes.c_int16], Generic[_P]):
    """Validator for 16-bit integers"""

    _size: ClassVar[int] = 2
    _unsigned: ClassVar[bool] = False
    _min: ClassVar[int] = -(2**15)
    _max: ClassVar[int] = 2**15 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_int16


class Int32(IntValidatorBase[_P, ctypes.c_int32], Generic[_P]):
    """Validator for 32-bit integers"""

    _size: ClassVar[int] = 4
    _unsigned: ClassVar[bool] = False
    _min: ClassVar[int] = -(2**31)
    _max: ClassVar[int] = 2**31 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_int32


class Int64(IntValidatorBase[_P, ctypes.c_int64], Generic[_P]):
    """Validator for 64-bit integers"""

    _size: ClassVar[int] = 8
    _unsigned: ClassVar[bool] = False
    _min: ClassVar[int] = -(2**63)
    _max: ClassVar[int] = 2**63 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_int64


class Uint8(IntValidatorBase[_P, ctypes.c_uint8], Generic[_P]):
    """Validator for unsigned 8-bit integers"""

    _size: ClassVar[int] = 1
    _unsigned: ClassVar[bool] = True
    _min: ClassVar[int] = 0
    _max: ClassVar[int] = 2**8 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_uint8


class Uint16(IntValidatorBase[_P, ctypes.c_uint16], Generic[_P]):
    """Validator for unsigned 16-bit integers"""

    _size: ClassVar[int] = 2
    _unsigned: ClassVar[bool] = True
    _min: ClassVar[int] = 0
    _max: ClassVar[int] = 2**16 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_uint16


class Uint32(IntValidatorBase[_P, ctypes.c_uint32], Generic[_P]):
    """Validator for unsigned 32-bit integers"""

    _size: ClassVar[int] = 4
    _unsigned: ClassVar[bool] = True
    _min: ClassVar[int] = 0
    _max: ClassVar[int] = 2**32 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_uint32


class Uint64(IntValidatorBase[_P, ctypes.c_uint64], Generic[_P]):
    """Validator for unsigned 64-bit integers"""

    _size: ClassVar[int] = 8
    _unsigned: ClassVar[bool] = True
    _min: ClassVar[int] = 0
    _max: ClassVar[int] = 2**64 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_uint64


class Byte(FieldValidator[_P, int], Generic[_P]):
    """Validator for single byte values"""

    _size: ClassVar[int] = 1
    _unsigned: ClassVar[bool] = True
    _min: ClassVar[int] = 0
    _max: ClassVar[int] = 2**8 - 1

    def __init__(self, *args) -> None:
        self._ctype: Type[ctypes._SimpleCData] = ctypes.c_ubyte

    @property
    def size(self) -> int:
        return self._size

    @property
    def unsigned(self) -> bool:
        return self._unsigned

    @property
    def min(self) -> int:
        return self._min

    @property
    def max(self) -> int:
        return self._max

    def __get__(self, obj: _P, objtype=None) -> int:
        return getattr(obj, self._private_name)

    def __set__(self, obj: _P, value: Union[int, bytes, bytearray, ctypes.c_ubyte]):
        if _VALIDATION_ENABLED.get():
            self.validate_one(value)
            if isinstance(value, (bytes, bytearray)):
                int_value = int.from_bytes(value, "little")
                setattr(obj, self._private_name, int_value)
                return
        setattr(obj, self._private_name, value)

    def validate_one(self, value: Union[int, bytes, bytearray, ctypes.c_ubyte]):
        """validate a single byte value

        Args:
            value (Union[int, bytes, bytearray]): Byte value to validate

        Raises:
            ValueError: Value out of range
            TypeError: Wrong type
        """

        if isinstance(value, self._ctype):
            return

        if isinstance(value, int):
            if not self._min <= value <= self._max:
                raise ValueError(
                    f"Expected {value} to be in range of {self._min} to {self._max}"
                )
            return

        if not isinstance(value, (bytes, bytearray)):
            raise TypeError(f"Expected {value!r} to be bytes or int")

        if len(value) != 1:
            raise ValueError(f"Expected {value!r} to be no longer than 1")

    def validate_many(self, value: Union[Iterable[int], bytes, bytearray]):
        """Validate multiple byte values

        Args:
            value (Union[Iterable[int], bytes, bytearray]): Byte values to validate

        Raises:
            TypeError: Wrong type
            ValueError: Value out of range
        """
        if isinstance(value, (bytes, bytearray)):
            return

        # Commented this check to help performance
        if any(not isinstance(v, int) for v in value):
            raise TypeError(
                f"Expected {value} to be an int sequence, bytes, or bytearray."
            )

        if (max(value) > self._max) or (min(value) < self._min):
            raise ValueError(
                f"Expected {value} to be in range of {self._min} to {self._max}."
            )

    def __repr__(self):
        return f"Byte(len=1) at 0x{id(self):016X}"


class String(FieldValidator[_P, str], Generic[_P]):
    """Validator for strings (char arrays)"""

    def __init__(self, len: int):
        assert len > 1
        self.len = len
        self._ctype = ctypes.c_char * len

    def __get__(self, obj: _P, objtype=None) -> str:
        return getattr(obj, self._private_name).decode("ascii")

    def __set__(self, obj: _P, value: str):
        if isinstance(value, self._ctype):
            setattr(obj, self._private_name, value)
            return

        if _VALIDATION_ENABLED.get():
            self.validate_one(value)
        setattr(obj, self._private_name, value.encode("ascii"))

    def validate_one(self, value: str):
        """Validate a string value

        Args:
            value (str): String value

        Raises:
            TypeError: Wrong type
            ValueError: String exceeds max length
        """
        if not isinstance(value, str):
            raise TypeError(f"Expected {value} to be a str")

        if len(value) > (self.len - 1):
            raise ValueError(f'Expected "{value}" to be no longer than {self.len - 1}')

        if not value.isascii():
            raise TypeError(f"Expected {value} to only contain valid ascii points")

    def validate_many(self, value):
        """Validate multiple strings

        Not implemented

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def __repr__(self):
        return f"String(len={self.len}) at 0x{id(self):016X}"


_FV = TypeVar("_FV", bound=FieldValidator)


class Char(String[_P], Generic[_P]):
    """Validator for scalar char values"""

    def __init__(self) -> None:
        self._ctype = ctypes.c_char
        self.len = 1

    def __set__(self, obj: _P, value: Union[str, ctypes.c_char]):
        if isinstance(value, ctypes.c_char):
            setattr(obj, self._private_name, value)
        else:
            super().__set__(obj, value)

    def validate_one(self, value: Union[str, ctypes.c_char]):
        """Validate a char value

        Args:
            value (str): String value

        Raises:
            TypeError: Wrong type
            ValueError: String exceeds max length
        """

        if isinstance(value, self._ctype):
            return

        if not isinstance(value, str):
            raise TypeError(f"Expected {value} to be a str")

        if len(value) > self.len:
            raise ValueError(f'Expected "{value}" to be no longer than {self.len}')

        if not value.isascii():
            raise TypeError(f"Expected {value} to only contain valid ascii points")

    def __repr__(self):
        return f"Char() at 0x{id(self):016X}"


class ArrayField(FieldValidator, abc.Sequence, Generic[_FV]):
    """Array field validator base class"""

    def __init__(self, validator: Type[_FV], len: int):
        """Array field validator base class

        Args:
            validator (Type[_FV]): Field validator class for datatype
            len (int): Field length
        """
        self._validator: FieldValidator = validator()
        self._len = len
        self._bound_obj: Optional[MessageBase] = None
        self._ctype = self._validator._ctype * len

    @classmethod
    def _bound(cls, obj: ArrayField[_FV], bound_obj: MessageBase) -> ArrayField[_FV]:
        new_obj = super().__new__(cls)
        new_obj._bound_obj = bound_obj
        new_obj._validator = obj._validator
        new_obj._len = obj._len
        new_obj._ctype = obj._ctype
        new_obj._owner = obj._owner
        new_obj._public_name = obj._public_name
        new_obj._private_name = obj._private_name

        return new_obj

    def __get__(self, obj: MessageBase, objtype=None) -> ArrayField[_FV]:
        """Return an Array bound to a message obj instance."""
        return ArrayField._bound(self, obj)

    def __set__(self, obj: MessageBase, value: Union[ArrayField[_FV], Sequence]):
        if isinstance(value, ArrayField):
            if _VALIDATION_ENABLED.get():
                self.validate_array(value)
            setattr(
                obj, self._private_name, getattr(value._bound_obj, value._private_name)
            )
        else:
            self.__get__(obj).__setitem__(slice(None), value)

    def __getitem__(self, key):
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")

        return getattr(self._bound_obj, self._private_name)[key]

    def __iter__(self) -> Iterator:
        return iter(self[:])

    def __setitem__(self, key, value):
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")

        if _VALIDATION_ENABLED.get():
            if isinstance(value, abc.Iterable) or hasattr(value, "__getitem__"):
                self.validate_many(value)
            else:
                self.validate_one(value)

        getattr(self._bound_obj, self._private_name)[key] = value

    def __len__(self) -> int:
        return self._len

    def __repr__(self) -> str:
        return f"ArrayField({type(self._validator).__name__}, len={self._len}) at 0x{id(self):016X}"

    def __str__(self):
        if self._bound_obj:
            return str(getattr(self._bound_obj, self._private_name)[:])
        else:
            return self.__repr__()

    def __eq__(self, value) -> bool:
        if not isinstance(value, abc.Sequence):
            return False
        if len(value) != self._len:
            return False
        for self_val, comp_val in zip(self, value):
            if self_val != comp_val:
                return False
        return True

    def validate_one(self, value):
        """Validate one value

        Args:
            value: Value to validate
        """
        self._validator.validate_one(value)

    def validate_many(self, value):
        """Validate multiple values

        Args:
            value: Values to validate
        """
        self._validator.validate_many(value)

    def validate_array(self, value: ArrayField):
        """Validate array

        Args:
            value (ArrayField): Array value to validate

        Raises:
            TypeError: Wrong type
        """
        if not isinstance(value, type(self)):
            raise TypeError(
                f"Expected a {self.__class__.__name__}({type(self._validator).__name__}). Got {type(value).__name__}."
            )

        if not isinstance(value._validator, type(self._validator)):
            raise TypeError(
                f"Expected an {self.__class__.__name__}({type(self._validator).__name__}, {len(self)}). Got {type(value).__name__}({type(value._validator).__name__}, {len(value)})."
            )

        if len(value) != len(self):
            raise ValueError(
                f"Array size mismatch. Expected an {self.__class__.__name__}({type(self._validator).__name__}, {len(self)}). Got {type(value).__name__}({type(value._validator).__name__}, {len(value)})."
            )

        if value._bound_obj is None:
            raise ValueError(
                f"The instance of {type(value).__name__}({type(value._validator).__name__}, {len(value)}) is not bound to a MessageBase object."
            )

        return


_IV = TypeVar("_IV", bound=IntValidatorBase)


class IntArray(ArrayField[_IV], Generic[_IV]):
    """IntArray validator class"""

    def __init__(self, validator: Type[_IV], len: int):
        """IntArray validator class

        Args:
            validator (Type[IV]): Field validator class for Int type
            len (int): Field length
        """
        self._validator = validator()
        self._len = len
        self._bound_obj: Optional[MessageBase] = None
        self._ctype = self._validator._ctype * len

    @classmethod
    def _bound(cls, obj: ArrayField[_IV], bound_obj: MessageBase) -> IntArray[_IV]:
        new_obj = super().__new__(cls)
        new_obj._bound_obj = bound_obj
        new_obj._validator = obj._validator
        new_obj._len = obj._len
        new_obj._ctype = obj._ctype
        new_obj._owner = obj._owner
        new_obj._public_name = obj._public_name
        new_obj._private_name = obj._private_name

        return new_obj

    def __get__(self, obj: MessageBase, objtype=None) -> IntArray[_IV]:
        """Return an Array bound to a message obj instance."""
        return IntArray._bound(self, obj)

    def __set__(
        self,
        obj: MessageBase,
        value: Union[ArrayField[_IV], Sequence[int], ctypes.Array],
    ):
        if isinstance(value, ArrayField):
            if _VALIDATION_ENABLED.get():
                self.validate_array(value)
            setattr(
                obj, self._private_name, getattr(value._bound_obj, value._private_name)
            )
        else:
            self.__get__(obj).__setitem__(slice(None), value)

    @overload
    def __getitem__(self, key: int) -> int: ...

    @overload
    def __getitem__(self, key: slice) -> List[int]: ...

    def __getitem__(self, key) -> Union[int, List[int]]:
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")

        return getattr(self._bound_obj, self._private_name)[key]

    def __iter__(self) -> Iterator[int]:
        return iter(self[:])

    def __repr__(self) -> str:
        return f"IntArray({type(self._validator).__name__}, len={self._len}) at 0x{id(self):016X}"


_FPV = TypeVar("_FPV", bound=FloatValidatorBase)


class ByteArray(ArrayField[Byte]):
    """Validator class for Bytes arrays"""

    def __init__(self, len: int):
        """Validator class for Bytes arrays

        Args:
            len (int): Byte array length
        """
        assert len > 1
        self._validator = Byte()
        self._len = len
        self._bound_obj: Optional[MessageBase] = None
        self._ctype = ctypes.c_ubyte * len

    @classmethod
    def _bound(cls, obj: ArrayField[Byte], bound_obj: MessageBase) -> ByteArray:
        new_obj = super().__new__(cls)
        new_obj._bound_obj = bound_obj
        new_obj._validator = obj._validator
        new_obj._len = obj._len
        new_obj._ctype = obj._ctype
        new_obj._owner = obj._owner
        new_obj._public_name = obj._public_name
        new_obj._private_name = obj._private_name

        return new_obj

    def __get__(self, obj: MessageBase, objtype=None) -> ByteArray:
        return ByteArray._bound(self, obj)

    def __set__(
        self,
        obj: MessageBase,
        value: Union[ArrayField[Byte], Sequence[int], bytes, bytearray, ctypes.Array],
    ):
        if isinstance(value, ArrayField):
            if _VALIDATION_ENABLED.get():
                self.validate_array(value)
            setattr(
                obj, self._private_name, getattr(value._bound_obj, value._private_name)
            )
        else:
            self.__get__(obj).__setitem__(slice(None), value)

    @overload
    def __getitem__(self, key: int) -> bytearray: ...

    @overload
    def __getitem__(self, key: slice) -> bytearray: ...

    def __getitem__(self, key) -> bytearray:
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")

        int_vals = getattr(self._bound_obj, self._private_name)[key]
        try:
            barray = [x.to_bytes(1, "little") for x in int_vals]
        except TypeError:
            barray = [int_vals.to_bytes(1, "little")]

        return bytearray().join(barray)

    def __iter__(self) -> Iterator[bytearray]:  # Generator[_S, None, None]:
        for i in range(self._len):
            yield self.__getitem__(i)

    def __setitem__(self, key, value):
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")

        if _VALIDATION_ENABLED.get():
            if isinstance(value, abc.Iterable) or hasattr(value, "__getitem__"):
                self.validate_many(value)
            else:
                self.validate_one(value)

            if isinstance(value, (bytes, bytearray)):
                if len(value) == 1:
                    value = int.from_bytes(value, "little")
                else:
                    value = [v for v in value]

        getattr(self._bound_obj, self._private_name)[key] = value

    def __repr__(self) -> str:
        return f"ByteArray(len={self._len}) at 0x{id(self):016X}"


class FloatArray(ArrayField[_FPV], Generic[_FPV]):
    """Validator class for float arrays"""

    def __init__(self, validator: Type[_FPV], len: int):
        """Validator class for float arrays

        Args:
            validator (Type[_FPV]): Float type validator
            len (int): Array length
        """
        self._validator = validator()
        self._len = len
        self._bound_obj: Optional[MessageBase] = None
        self._ctype = self._validator._ctype * len

    @classmethod
    def _bound(cls, obj: ArrayField[_FPV], bound_obj: MessageBase) -> FloatArray[_FPV]:
        new_obj = super().__new__(cls)
        new_obj._bound_obj = bound_obj
        new_obj._validator = obj._validator
        new_obj._len = obj._len
        new_obj._ctype = obj._ctype
        new_obj._owner = obj._owner
        new_obj._public_name = obj._public_name
        new_obj._private_name = obj._private_name

        return new_obj

    def __get__(self, obj: MessageBase, objtype=None) -> FloatArray[_FPV]:
        """Return an Array bound to a message obj instance."""
        return FloatArray._bound(self, obj)

    def __set__(
        self,
        obj: MessageBase,
        value: Union[ArrayField[_FPV], Sequence[float], ctypes.Array],
    ):
        if isinstance(value, ArrayField):
            if _VALIDATION_ENABLED.get():
                self.validate_array(value)
            setattr(
                obj, self._private_name, getattr(value._bound_obj, value._private_name)
            )
        else:
            self.__get__(obj).__setitem__(slice(None), value)

    @overload
    def __getitem__(self, key: int) -> float: ...

    @overload
    def __getitem__(self, key: slice) -> List[float]: ...

    def __getitem__(self, key) -> Union[float, List[float]]:
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")

        return getattr(self._bound_obj, self._private_name)[key]

    def __iter__(self) -> Iterator[float]:
        return iter(self[:])

    def __repr__(self) -> str:
        return f"FloatArray({type(self._validator).__name__}, len={self._len}) at 0x{id(self):016X}"


_S = TypeVar("_S", bound=MessageBase)


class Struct(FieldValidator, Generic[_S]):
    """Validator class for Structures"""

    def __init__(self, _ctype: Type[_S]):
        self._ctype = _ctype

    def __get__(self, obj, objtype=None) -> _S:
        return getattr(obj, self._private_name)

    @overload
    def __set__(self, obj: MessageBase, value: _S): ...

    @overload
    def __set__(self, obj: StructArray, value: Struct[_S]): ...

    def __set__(self, obj, value):
        if _VALIDATION_ENABLED.get():
            self.validate_one(value)
        # Note: ctypes already copies the data here
        setattr(obj, self._private_name, value)

    def validate_one(self, value: _S):
        """Validate a structure

        Args:
            value (_S): Structure value to validate

        Raises:
            TypeError: Wrong type
        """
        if not isinstance(value, self._ctype):
            raise TypeError(f"Expected {self._ctype.__name__}")

    def validate_many(self, value: Iterable[_S]):
        """Validate multiple structures

        Args:
            value (Iterable[_S]): Iterable of structures to validate

        Raises:
            TypeError: Wrong type
        """
        if any(not isinstance(v, self._ctype) for v in value):
            raise TypeError(f"Expected {value} to be an {self._ctype.__name__}.")

    def __repr__(self) -> str:
        return f"Struct({self._ctype.__name__}) at 0x{id(self):016X}"


class StructArray(FieldValidator, abc.Sequence, Generic[_S]):
    """Validator for structure arrays"""

    def __init__(self, msg_struct: Type[_S], len: int):
        """Validator for structure arrays

        Args:
            msg_struct (Type[_S]): Structure class
            len (int): Array length
        """
        self._validator = Struct(msg_struct)
        self._len = len
        self._bound_obj: Optional[MessageBase] = None
        self._ctype = self._validator._ctype * len

    @classmethod
    def _bound(cls, obj: StructArray[_S], bound_obj: MessageBase) -> StructArray[_S]:
        new_obj: StructArray[_S] = super().__new__(cls)
        new_obj._bound_obj = bound_obj
        new_obj._validator = obj._validator
        new_obj._len = obj._len
        new_obj._ctype = obj._ctype
        new_obj._owner = obj._owner
        new_obj._public_name = obj._public_name
        new_obj._private_name = obj._private_name

        return new_obj

    def __get__(self, obj: MessageBase, objtype=None) -> StructArray[_S]:
        """Return an StructArray bound to a message obj instance."""
        return StructArray._bound(self, obj)

    def __set__(self, obj, value: Union[StructArray, Sequence[_S], ctypes.Array[_S]]):
        if isinstance(value, StructArray):
            if _VALIDATION_ENABLED.get():
                self.validate_array(value)
            setattr(
                obj, self._private_name, getattr(value._bound_obj, value._private_name)
            )
        else:
            self.__get__(obj).__setitem__(slice(None), value)

    @overload
    def __getitem__(self, key: int) -> _S: ...

    @overload
    def __getitem__(self, key: slice) -> List[_S]: ...

    def __getitem__(self, key) -> Union[_S, List[_S]]:
        if self._bound_obj is None:
            raise AttributeError("Array descriptor is not bound to an instance object.")
        return getattr(self._bound_obj, self._private_name)[key]

    def __iter__(self) -> Iterator[_S]:
        return iter(self[:])

    def __setitem__(self, key, value):
        if self._bound_obj is None:
            raise AttributeError(
                "StructArray descriptor is not bound to an instance object."
            )

        if _VALIDATION_ENABLED.get():
            if isinstance(value, abc.Iterable) or hasattr(value, "__getitem__"):
                self.validate_many(value)
            else:
                self.validate_one(value)

        getattr(self._bound_obj, self._private_name)[key] = value

    def __len__(self) -> int:
        return self._len

    def __repr__(self) -> str:
        return f"StructArray({self._validator._ctype.__name__}, len={self._len}) at 0x{id(self):016X}"

    def __str__(self):
        return self.pretty_print()

    def pretty_print(self, add_tabs=0):
        """Generate formatted string for structure array

        Args:
            add_tabs (int, optional): Indent level. Defaults to 0.

        Returns:
            str: Formatted string
        """
        if self._bound_obj:
            max_len = 5
            val = getattr(self._bound_obj, self._private_name)
            tab = "\t" * add_tabs
            if self._len <= max_len:
                return (
                    f"{tab}[\n"
                    + "\n".join(x.pretty_print(add_tabs) for x in val)
                    + f"\n{tab}]"
                )
            else:
                return (
                    f"{tab}[\n"
                    + "\n".join(
                        f"{val[i].pretty_print(add_tabs)}" for i in range(max_len)
                    )
                    + f"\n{tab}...]"
                )
        else:
            return self.__repr__()

    def __eq__(self, value) -> bool:
        if not isinstance(value, abc.Sequence):
            return False
        if len(value) != self._len:
            return False
        for self_val, comp_val in zip(self, value):
            if self_val != comp_val:
                return False
        return True

    def validate_one(self, value: _S):
        """Validate a structure

        Args:
            value (_S): Structure value to validate
        """
        self._validator.validate_one(value)

    def validate_many(self, value: Iterable[_S]):
        """Validate multiple structures

        Args:
            value (Iterable[_S]): Structure values to validate
        """
        self._validator.validate_many(value)

    def validate_array(self, value: StructArray[_S]):
        """Validate structure array

        Args:
            value (StructArray[_S]): StructArray to validate

        Raises:
            TypeError: Wrong type
        """
        if not isinstance(value, StructArray):
            raise TypeError(
                f"Expected a StructArray({self._validator._ctype.__name__}, {len(self)}). Got {type(value).__name__}."
            )

        if value._validator._ctype is not self._validator._ctype:
            raise TypeError(
                f"Expected a StructArray({self._validator._ctype.__name__}, {len(self)}). Got {type(value).__name__}({value._validator._ctype.__name__}, {len(value)})."
            )

        if len(value) != len(self):
            raise ValueError(
                f"Array size mismatch. Expected an {self.__class__.__name__}({self._validator._ctype.__name__}, {len(self)}). Got {type(value).__name__}({value._validator._ctype.__name__}, {len(value)})."
            )

        if value._bound_obj is None:
            raise ValueError(
                f"The instance of {type(value).__name__}({value._validator._ctype.__name__}, {len(value)}) is not bound to a MessageBase object."
            )

        return



import pathlib

from textwrap import dedent
from pyrtma.parser import (
    Parser,
    ConstantExpr,
    ConstantString,
    TypeAlias,
    MT,
    MID,
    HID,
    MDF,
    SDF,
)
from typing import Union
import subprocess
import sys
import os
from pyrtma.__version__ import __version__


# Field type name to ctypes
type_map = {
    "char": "ctypes.c_char",
    "unsigned char": "ctypes.c_ubyte",
    "byte": "ctypes.c_ubyte",
    "float": "ctypes.c_float",
    "double": "ctypes.c_double",
    "uint8": "ctypes.c_uint8",
    "uint16": "ctypes.c_uint16",
    "uint32": "ctypes.c_uint32",
    "uint64": "ctypes.c_uint64",
    "int8": "ctypes.c_int8",
    "int16": "ctypes.c_int16",
    "int32": "ctypes.c_int32",
    "int64": "ctypes.c_int64",
    # following unsized types are deprecated
    "int": "ctypes.c_int32",
    "signed int": "ctypes.c_int32",
    "unsigned int": "ctypes.c_uint32",
    "unsigned": "ctypes.c_uint32",
    "short": "ctypes.c_int16",
    "signed short": "ctypes.c_int16",
    "unsigned short": "ctypes.c_uint16",
    "long": "ctypes.c_int32",
    "signed long": "ctypes.c_int32",
    "unsigned long": "ctypes.c_uint32",
    "long long": "ctypes.c_int64",
    "signed long long": "ctypes.c_int64",
    "unsigned long long": "ctypes.c_uint64",
}

desctype_map = {
    "char": "Char",
    "byte": "Byte",
    "unsigned char": "Byte",
    "float": "Float",
    "double": "Double",
    "uint8": "Uint8",
    "uint16": "Uint16",
    "uint32": "Uint32",
    "uint64": "Uint64",
    "int8": "Int8",
    "int16": "Int16",
    "int32": "Int32",
    "int64": "Int64",
    # following unsized types are deprecated
    "int": "Int32",
    "signed int": "Int32",
    "unsigned int": "Uint32",
    "unsigned": "Uint32",
    "short": "Int16",
    "signed short": "Int16",
    "unsigned short": "Uint16",
    "long": "Int32",
    "signed long": "Int32",
    "unsigned long": "Uint32",
    "long long": "Int64",
    "signed long long": "Int64",
    "unsigned long long": "Uint64",
}
DEFAULT_BLACK_LEN = 88  # preferred line length
TAB = "    "


class PyDefCompiler:
    def __init__(self, parser: Parser, debug: bool = False):
        self.debug = debug
        self.parser = parser

    def generate_constant(self, c: ConstantExpr) -> str:
        return f"{c.name}: {type(c.value).__qualname__} = {c.value}\n"

    def generate_string_constant(self, c: ConstantString) -> str:
        return f"{c.name}: str = {c.value}\n"

    def generate_msg_type_id(self, mt: MT) -> str:
        return f"MT_{mt.name}: int = {mt.value}\n"

    def generate_module_id(self, mid: MID) -> str:
        return f"MID_{mid.name}: int = {mid.value}\n"

    def generate_host_id(self, hid: HID) -> str:
        return f"{hid.name}: int = {hid.value}\n"

    def generate_type_alias(self, td: TypeAlias) -> str:
        ftype = type_map.get(td.type_name)
        if ftype:
            return f"{td.name} = {ftype}\n"
        else:
            return f"{td.name} = {td.type_name}\n"

    def get_descriptor(self, ftype: str, flen: int) -> str:
        if ftype in type_map.keys():
            dtype = desctype_map[ftype]
            # Scalar Field
            if flen <= 1:
                return f":{dtype} = {dtype}()"

            # Array Field
            if dtype.startswith("Int") or dtype.startswith("Uint"):
                return f":IntArray[{dtype}] = IntArray({dtype}, {flen})"
            elif dtype.startswith("Float") or dtype.startswith("Double"):
                return f":FloatArray[{dtype}] = FloatArray({dtype}, {flen})"
            elif dtype.startswith("Char"):
                return f":String = String({flen})"
            elif dtype.startswith("Byte"):
                return f":ByteArray = ByteArray({flen})"
            else:
                raise RuntimeError(f"Unknown field descriptor {dtype}")

        # Message Def (Struct) Field
        elif ftype in self.parser.message_defs.keys():
            ftype = f"MDF_{ftype}"
            return (
                f":Struct[{ftype}] = Struct({ftype})"
                if flen == 0
                else f":StructArray[{ftype}] = StructArray({ftype}, {flen})"
            )

        # Struct Field
        elif ftype in self.parser.struct_defs.keys():
            ftype = f"{ftype}"
            return (
                f":Struct[{ftype}] = Struct({ftype})"
                if flen == 0
                else f":StructArray[{ftype}] = StructArray({ftype}, {flen})"
            )

        # Alias Field
        elif ftype in self.parser.aliases.keys():
            ftype = self.parser.aliases[ftype].type_name
            return self.get_descriptor(ftype, flen)

        else:
            raise RuntimeError(f"Unknown field name {ftype}")

    def generate_struct(self, sdf: SDF) -> str:
        dstr = "\n"
        for field in sdf.fields:
            desc = self.get_descriptor(field.type_name, field.length or 0)
            dstr += f"{'    ' * 3}{field.name}{desc}\n"

        msg_src = sdf.src.as_posix()
        type_def_str = repr(sdf.raw)
        type_def_rhs = 'type_def: ClassVar[str] = "'
        type_def_end = '"'
        type_def_line = f"{type_def_rhs}{type_def_str}{type_def_end}"
        type_def_line_len = len(TAB) + len(type_def_line)
        if type_def_line_len > DEFAULT_BLACK_LEN:
            type_def_rhs = f'type_def: ClassVar[\n{TAB *4}str\n{TAB *3}] = "'
            type_def_line = f"{type_def_rhs}{type_def_str}{type_def_end}"

        template = f"""\
        class {sdf.name}(MessageBase, metaclass=MessageMeta):
            type_name: ClassVar[str] = \"{sdf.name}\"
            type_hash: ClassVar[int] = 0x{sdf.hash[:8].upper()}
            type_size: ClassVar[int] = {sdf.size}
            type_source: ClassVar[str] = \"{msg_src}\"
            {type_def_line}

            {dstr}
        """
        return dedent(template)

    def generate_msg_def(self, mdf: MDF) -> str:
        dstr = "\n"
        for field in mdf.fields:
            desc = self.get_descriptor(field.type_name, field.length or 0)
            dstr += f"{'    ' * 3}{field.name}{desc}\n"

        msg_id = mdf.type_id
        msg_src = mdf.src.as_posix()
        type_def_str = repr(mdf.raw)
        type_def_rhs = 'type_def: ClassVar[str] = "'
        type_def_end = '"'
        type_def_line = f"{type_def_rhs}{type_def_str}{type_def_end}"
        type_def_line_len = len(TAB) + len(type_def_line)
        if type_def_line_len > DEFAULT_BLACK_LEN:
            type_def_rhs = f'type_def: ClassVar[\n{TAB *4}str\n{TAB *3}] = "'
            type_def_line = f"{type_def_rhs}{type_def_str}{type_def_end}"

        template = f"""\
        @pyrtma.message_def
        class MDF_{mdf.name}(MessageData, metaclass=MessageMeta):
            type_id: ClassVar[int] = {msg_id}
            type_name: ClassVar[str] = \"{mdf.name}\"
            type_hash: ClassVar[int] = 0x{mdf.hash[:8].upper()}
            type_size: ClassVar[int] = {mdf.size}
            type_source: ClassVar[str] = \"{msg_src}\"
            {type_def_line}

            {dstr}
        """
        return dedent(template)

    def generate_docstr(self) -> str:
        s = f"""\"\"\"This message def file was auto-generated by pyrtma.compile version {__version__}\"\"\"\n"""
        return s

    def generate_imports(self) -> str:
        s = """\
        import ctypes

        import pyrtma
        from pyrtma.__version__ import check_compiled_version, __version__
        from packaging.version import parse as _ver_parse
        from typing import ClassVar

        from pyrtma.message_base import MessageBase, MessageMeta
        from pyrtma.message_data import MessageData
        from pyrtma.validators import Int8, Int16, Int32, Int64, Uint8, Uint16, Uint32, Uint64, Float, Double, Struct, IntArray, FloatArray, StructArray, Char, String, Byte, ByteArray
        
        if _ver_parse(__version__) >= _ver_parse("2.3.0"):
            from pyrtma.context import _update_context, get_context

        """
        return dedent(s)

    def generate_context(self):
        s = """\
        if _ver_parse(__version__) >= _ver_parse("2.3.0"):
            _update_context(__name__)
        """
        return dedent(s)

    def generate(self, out_filepath: pathlib.Path):
        with open(out_filepath, mode="w") as f:
            f.write(self.generate_docstr())
            f.write(self.generate_imports())
            f.write("\n")

            # Version
            f.write(f'COMPILED_PYRTMA_VERSION: str = "{__version__}"\n')
            f.write("check_compiled_version(COMPILED_PYRTMA_VERSION)\n\n")

            f.write("# Constants\n")
            obj: Union[ConstantExpr, ConstantString, HID, MID, TypeAlias, SDF, MT, MDF]
            for obj in self.parser.constants.values():
                f.write(self.generate_constant(obj))
            f.write("\n")

            f.write("# String Constants\n")
            for obj in self.parser.string_constants.values():
                f.write(self.generate_string_constant(obj))
            f.write("\n")

            f.write("# Type Aliases\n")
            for obj in self.parser.aliases.values():
                f.write(self.generate_type_alias(obj))
            f.write("\n")

            f.write("# Host IDs\n")
            for obj in self.parser.host_ids.values():
                f.write(self.generate_host_id(obj))
            f.write("\n")

            f.write("# Module IDs\n")
            for obj in self.parser.module_ids.values():
                f.write(self.generate_module_id(obj))
            f.write("\n")

            f.write("# Message Type IDs\n")
            for obj in self.parser.message_ids.values():
                f.write(self.generate_msg_type_id(obj))
            f.write("\n\n")

            f.write("# Struct Definitions\n")
            for obj in self.parser.struct_defs.values():
                f.write(self.generate_struct(obj))
                f.write("\n\n")

            f.write("# Message Definitions\n")
            for obj in self.parser.message_defs.values():
                f.write(self.generate_msg_def(obj))
                f.write("\n\n")

            f.write("# User Context\n")
            f.write(self.generate_context())

        # run black formatter
        subprocess.run([sys.executable, "-m", "black", out_filepath], cwd=os.getcwd())



import ctypes


def print_ctype_array(arr: ctypes.Array):
    """Expand and print ctype arrays to string

    Args:
        arr (ctypes.Array): ctype array

    Returns:
        str: string representation of ctype array
    """
    """expand and print ctype arrays"""
    max_len = 20
    arr_len = len(arr)
    str = "["
    for i in range(0, min(arr_len, max_len)):
        val = arr[i]
        if isinstance(val, ctypes.Array):
            val = print_ctype_array(val)
        str += f"{val}, "
    if arr_len > max_len:
        str += "...]"
    else:
        str = str[:-2] + "]"
    return str


def hexdump(src: bytes, length=16, sep=" "):
    """Dump bytes as hex

    Args:
        src (bytes): bytes to hexdump
        length (int, optional): Row length. Defaults to 16.
        sep (str, optional): Separator for non-printable characters. Defaults to " ".
    """
    FILTER = "".join([(len(repr(chr(x))) == 3) and chr(x) or sep for x in range(256)])
    for c in range(0, len(src), length):
        chars = src[c : c + length]
        hex_ = " ".join(["{:02x}".format(x) for x in chars])
        if len(hex_) > 24:
            hex_ = "{} {}".format(hex_[:24], hex_[24:])
        printable = "".join(
            ["{}".format((x <= 127 and FILTER[x]) or sep) for x in chars]
        )
        print(
            "{0:08x}  {1:{2}s} |{3:{4}s}|".format(
                c, hex_, length * 3, printable, length
            )
        )



from sys import version_info
import unittest

if version_info[:2] < (3, 8):
    raise unittest.SkipTest("parser does not support < 3.8")

import tempfile
import pyrtma.parser
import textwrap
import pathlib
import logging
import unittest


class TempDefFile:
    def __init__(self):
        self.file = tempfile.NamedTemporaryFile("w", suffix=".yaml", delete=False)
        self.file.close()
        self.path = pathlib.Path(self.file.name)

    def __del__(self):
        if not self.file.closed:
            self.file.close()

        self.path.unlink()

    def write(self, text: str):
        if not self.file.closed:
            self.file.close()

        with open(self.path, "w") as f:
            f.write(text)


class TestParser(unittest.TestCase):
    def setUp(self) -> None:
        self.parser = pyrtma.parser.Parser()
        self.parser.logger.setLevel(logging.ERROR)

        self.tmp = TempDefFile()
        self.tmp2 = TempDefFile()

    def TearDown(self) -> None:
        del self.tmp
        del self.tmp2

    def test_host_id_conflicts(self):
        dup_host_id = textwrap.dedent(
            """
            host_ids:
                A: 10
                B: 10
            """
        )

        self.tmp.write(dup_host_id)

        with self.assertRaises(pyrtma.parser.HostIDError):
            self.parser.parse(self.tmp.path)

    def test_module_id_conflicts(self):
        dup_mod_id = textwrap.dedent(
            """
            module_ids:
                A: 10
                B: 10
            """
        )

        self.tmp.write(dup_mod_id)

        with self.assertRaises(pyrtma.parser.ModuleIDError):
            self.parser.parse(self.tmp.path)

    def test_signal_id_conflicts(self):
        # Duplicate Signal
        dup_msg_id = textwrap.dedent(
            """
            message_defs:
                A:
                    id: 100
                    fields: null
                B:
                    id: 100
                    fields: null
            """
        )

        self.tmp.write(dup_msg_id)

        with self.assertRaises(pyrtma.parser.MessageIDError):
            self.parser.parse(self.tmp.path)

    def test_message_id_conflicts(self):
        # Duplicate Message
        dup_msg_id = textwrap.dedent(
            """
            message_defs:
                A:
                    id: 100
                    fields:
                        i: int
                B:
                    id: 100
                    fields:
                        i: int
            """
        )

        self.tmp.write(dup_msg_id)

        with self.assertRaises(pyrtma.parser.MessageIDError):
            self.parser.parse(self.tmp.path)

    def test_duplicate_message(self):
        dup_msg = textwrap.dedent(
            """
            message_defs:
                A:
                    id: 100
                    fields:
                        i: int
            """
        )

        self.tmp.write(dup_msg)
        self.parser.parse(self.tmp.path)
        self.tmp2.write(dup_msg)

        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp2.path)

    def test_duplicate_module(self):
        dup_msg = textwrap.dedent(
            """
            module_ids:
                A: 1000
            """
        )

        self.tmp.write(dup_msg)
        self.parser.parse(self.tmp.path)
        self.tmp2.write(dup_msg)

        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp2.path)

    def test_duplicate_host(self):
        dup_msg = textwrap.dedent(
            """
            host_ids:
                A: 80
            """
        )

        self.tmp.write(dup_msg)
        self.parser.parse(self.tmp.path)
        self.tmp2.write(dup_msg)

        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp2.path)

    def test_duplicate_constant(self):
        dup_msg = textwrap.dedent(
            """
            constants:
                A: 80
            """
        )

        self.tmp.write(dup_msg)
        self.parser.parse(self.tmp.path)
        self.tmp2.write(dup_msg)

        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp2.path)

    def test_constant_name_conflicts(self):
        text = textwrap.dedent(
            """
            constants:
                A: 80
            string_constants:
                A: "abc"
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            constants:
                A: 80
            string_constants:
                A: "abc"
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            constants:
                A: 80
            aliases:
                A: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            constants:
                A: 80
            struct_defs:
                A:
                    id: 1000
                    fields: null
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            constants:
                A: 80
            message_defs:
                A:
                    id: 1000
                    fields: null
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

    def test_string_constant_name_conflicts(self):
        text = textwrap.dedent(
            """
            constants:
                A: 80
            string_constants:
                A: "abc"
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            string_constants:
                A: "abc"
            aliases:
                A: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            string_constants:
                A: "abc"
            struct_defs:
                A:
                    id: 1000
                    fields: null
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            string_constants:
                A: "abc"
            message_defs:
                A:
                    id: 1000
                    fields: null
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

    def test_alias_name_conflicts(self):
        text = textwrap.dedent(
            """
            constants:
                A: 80
            aliases:
                A: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            string_constants:
                A: "abc"
            aliases:
                A: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            aliases:
                A: int
            struct_defs:
                A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            aliases:
                A: int
            message_defs:
                A:
                    id: 1000
                    fields: null
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

    def test_struct_name_conflicts(self):
        text = textwrap.dedent(
            """
            constants:
                A: 80
            struct_defs:
                A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            string_constants:
                A: "abc"
            struct_defs:
                A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            aliases:
                A: int
            struct_defs:
                A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            struct_defs:
                A:
                    fields:
                        i: int
            message_defs:
                A:
                    id: 1000
                    fields: null
            """
        )

        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

    def test_message_name_conflicts(self):
        text = textwrap.dedent(
            """
            constants:
                A: 80
            message_defs:
                A:
                    id: 1000
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            string_constants:
                A: "abc"
            message_defs:
                A:
                    id: 1000
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            aliases:
                A: int
            message_defs:
                A:
                    id: 1000
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            struct_defs:
                A:
                    fields:
                        i: int
            message_defs:
                A:
                    id: 1000
                    fields: null
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.DuplicateNameError):
            self.parser.parse(self.tmp.path)

    def test_yaml_syntax(self):
        # Duplicate Key
        text = textwrap.dedent(
            """
            struct_defs:
                A:
                    fields:
                        i: int
            struct_defs:
                A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.YAMLSyntaxError):
            self.parser.parse(self.tmp.path)

    def test_rtma_syntax(self):
        # Bad Name
        text = textwrap.dedent(
            """
            struct_defs:
                _A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        # Bad Name
        text = textwrap.dedent(
            """
            message_defs:
                _A:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        # Bad Name
        text = textwrap.dedent(
            """
            constants:
                _A: 10
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        # Bad Name
        text = textwrap.dedent(
            """
            aliases:
                _A: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        # Illegal field name
        text = textwrap.dedent(
            """
            message_defs:
                A:
                    id:  1000
                    fields:
                        type_name: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        # Bad Indentation
        text = textwrap.dedent(
            """
            struct_defs:
            A:
                fields:
                    i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        # Wrong Type for imports
        text = textwrap.dedent(
            """
            imports:
                A: abc
            """
        )

        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

    def test_type_error(self):
        # Wrong Type for host_id
        text = textwrap.dedent(
            """
            host_ids:
                A: abc
            """
        )

        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.InvalidTypeError):
            self.parser.parse(self.tmp.path)

        # Wrong Type for module_id
        text = textwrap.dedent(
            """
            module_ids:
                A: abc
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.InvalidTypeError):
            self.parser.parse(self.tmp.path)

    def test_reserve_syntax(self):
        text = textwrap.dedent(
            """
            message_defs:
                _RESERVED_:
                    fields:
                        i: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            message_defs:
                _RESERVED_: 1
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.InvalidTypeError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            message_defs:
                _RESERVED_: [1:2:10]
            """
        )
        self.tmp.write(text)
        with self.assertRaises(
            (pyrtma.parser.YAMLSyntaxError, pyrtma.parser.InvalidTypeError)
        ):
            # YAMLSyntaxError does not happen on 3.13 (expression is read as string instead)
            # InvalidTypeError will happen if not caught by yaml loader
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            message_defs:
                _RESERVED_: [1-2-10]
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.InvalidTypeError):
            self.parser.parse(self.tmp.path)

        text = textwrap.dedent(
            """
            message_defs:
                _RESERVED_:
                    id: 1
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.InvalidTypeError):
            self.parser.parse(self.tmp.path)

    def test_auto_padding(self):
        text = textwrap.dedent(
            """
            message_defs:
                A:
                    id: 1001
                    fields:
                        i: int
                        d: double
                        c: char
                B:
                    id: 1002
                    fields:
                        c: char
                        i: int
                        d: double
                C:
                    id: 1003
                    fields:
                        f: float[2]
                        i: int64
                        d: int32[3]
                D:
                    id: 1004
                    fields:
                        f: float
                        i: int64
                        c: char[7]
                E:
                    id: 1005
                    fields:
                        a: A
                        b: B
                        c: C
                        d: D

            """
        )

        self.tmp.write(text)
        self.parser.parse(self.tmp.path)

    def test_reserved_fields(self):
        text = textwrap.dedent(
            """
            message_defs:
                _RESERVED_:
                    fields:
                        auto: int
            """
        )
        self.tmp.write(text)
        with self.assertRaises(pyrtma.parser.RTMASyntaxError):
            self.parser.parse(self.tmp.path)



def main():
    """Main function for starting base Data Logger"""
    import argparse
    import pathlib
    import sys
    import importlib
    import logging
    from ..exceptions import ConnectionLost, MessageManagerNotFound
    from .data_logger import DataLogger

    parser = argparse.ArgumentParser(description="Packet Data Logger")

    parser.add_argument(
        "-m",
        "--mm-ip",
        default="127.0.0.1:7111",
        dest="mm_ip",
        type=str,
        help="IP address of Message Manager. Defaults to 127.0.0.1:7111.",
    )

    parser.add_argument(
        "-d",
        "--defs",
        dest="defs_file",
        type=str,
        help="Path to python message definitions file. Required argument.",
    )

    parser.add_argument(
        "--log-level",
        dest="log_level",
        choices=["DEBUG", "INFO", "WARN", "ERROR"],
        default="INFO",
        help="Logging Level (defaults to INFO)",
    )

    args = parser.parse_args()

    if args.log_level == "DEBUG":
        level = logging.DEBUG
    elif args.log_level == "INFO":
        level = logging.INFO
    elif args.log_level == "WARN":
        level = logging.WARN
    elif args.log_level == "ERROR":
        level = logging.ERROR
    else:
        print("Unknown log level. Using INFO instead")
        level = logging.INFO

    if args.defs_file:
        base = pathlib.Path(args.defs_file).absolute().parent
        fname = pathlib.Path(args.defs_file).stem

        sys.path.insert(0, (str(base.absolute())))
        importlib.import_module(fname)

    d = DataLogger(args.mm_ip, log_level=level)

    try:
        d.run()
    except MessageManagerNotFound as e:
        d.mod.error(f"MessageManagerNotFound: {e.args[0]}")
        sys.exit(1)
    except ConnectionLost as e:
        d.mod.error(f"ConnectionLost: Connection to MessageManager is broken.")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""v1 Message Type Compiler: .h to .py"""

import re
from typing import List, Dict, Optional, Tuple, IO
import ctypes

# Field type name to ctypes
ctypes_map = {
    "char": ctypes.c_char,
    "unsigned char": ctypes.c_ubyte,
    "byte": ctypes.c_ubyte,
    "int": ctypes.c_int,
    "signed int": ctypes.c_uint,
    "unsigned int": ctypes.c_uint,
    "unsigned": ctypes.c_uint,
    "short": ctypes.c_short,
    "signed short": ctypes.c_short,
    "unsigned short": ctypes.c_ushort,
    "long": ctypes.c_long,
    "signed long": ctypes.c_long,
    "unsigned long": ctypes.c_ulong,
    "long long": ctypes.c_longlong,
    "signed long long": ctypes.c_longlong,
    "unsigned long long": ctypes.c_ulonglong,
    "float": ctypes.c_float,
    "double": ctypes.c_double,
    "int8_t": ctypes.c_int8,
    "int16_t": ctypes.c_int16,
    "int32_t": ctypes.c_int32,
    "int64_t": ctypes.c_int64,
    "uint8_t": ctypes.c_uint8,
    "uint16_t": ctypes.c_uint16,
    "uint32_t": ctypes.c_uint32,
    "uint64_t": ctypes.c_uint64,
    "MODULE_ID": ctypes.c_short,
    "HOST_ID": ctypes.c_short,
    "MSG_TYPE": ctypes.c_int,
    "MSG_COUNT": ctypes.c_int,
}


def camelcase(name):
    name = re.sub(r"_", " ", name)
    pieces = [s.title() for s in name.split(sep=" ")]
    return "".join(pieces)


def preprocess(text: str) -> str:
    # Strip Inline Comments
    text = re.sub(r"//(.*)\n", r"\n", text)

    # Strip Block Comments
    text = re.sub(r"/\*(.*?)\*/", r"\n", text, flags=re.DOTALL)

    # Strip Tabs
    text = re.sub(r"\t+", " ", text)

    return text


def parse_defines(text: str) -> Dict:
    defines: Dict[str, Dict[str, str]] = {}
    defines["MT"] = {}
    defines["MID"] = {}
    defines["constants"] = {}

    # Get Defines (Only simple one line constants here)
    macros: List[Tuple[str, str]] = re.findall(
        r"#define\s*(?P<name>\w+)\s+(?P<expression>[\(\)\[\]\w \*/\+-\.]+)\n", text
    )

    for name, exp in macros:
        name = name.strip()
        exp = exp.strip()

        # Store a mapping for each define type
        if name.startswith("MT_"):
            defines["MT"][name] = exp
        elif name.startswith("MID_"):
            defines["MID"][name] = exp
        else:
            if "/" in exp and "." not in exp:
                # Ensure division results in an integer if no decimal
                exp = "int(" + exp + ")"
            defines["constants"][name] = exp

    return defines


def parse_typedefs(text: str) -> Dict:
    # Get simple typedefs
    typedefs = {}
    c_typedefs = re.finditer(
        r"\s*typedef\s+(?P<qual1>\w+\s+)?\s*(?P<qual2>\w+\s+)?\s*(?P<typ>\w+)\s+(?P<name>\w+)\s*;\s*$",
        text,
        flags=re.MULTILINE,
    )

    for m in c_typedefs:
        alias = m.group("name").strip()
        if alias.startswith("MDF_"):
            # TODO:Non struct message defintion. Maybe drop support?
            pass
        else:
            # Field type
            qual1 = m.group("qual1")
            qual2 = m.group("qual2")
            typ = m.group("typ")

            ftype = ""
            if qual1:
                ftype += qual1.strip()
                ftype += " "

            if qual2:
                ftype += qual2.strip()
                ftype += " "

            ftype += typ.strip()

            # Must be a native type
            ctype = ctypes_map[ftype]

            # Create another alias for the native type
            typedefs[alias.strip()] = f"{ctype.__module__}.{ctype.__name__}"

    return typedefs


def parse_structs(
    msg_types: Dict[str, str], text: str
) -> Dict[str, Optional[List[Tuple[str, str, Optional[str]]]]]:
    structs: Dict[str, Optional[List[Tuple[str, str, Optional[str]]]]] = {}

    # Strip Newlines
    text = re.sub(r"\n", "", text)
    # Get Struct Definitions
    c_msg_defs = re.finditer(
        r"\s*typedef\s+struct\s*\{(?P<def>.*?)\}(?P<name>\s*\w*)\s*;", text
    )

    for m in c_msg_defs:
        name: str = m.group("name").strip()
        fields = m.group("def").split(sep=";")
        fields = [f.strip() for f in fields if f.strip() != ""]
        c_fields: List[Tuple[str, str, Optional[str]]] = []

        for field in fields:
            fmatch = re.match(
                r"(?P<qual1>\w+\s+)?\s*(?P<qual2>\w+\s+)?\s*(?P<typ>\w+)\s+(?P<name>\w+)\s*(\[(?P<length>.*)\])?$",
                field,
            )

            if fmatch is None:
                print(field)
                raise RuntimeError("Error parsing field definition.")

            # Field name
            fname = fmatch.group("name").strip()
            qual1 = fmatch.group("qual1")
            qual2 = fmatch.group("qual2")
            typ = fmatch.group("typ")

            ftype = ""
            if qual1:
                ftype += qual1.strip()
                ftype += " "

            if qual2:
                ftype += qual2.strip()
                ftype += " "

            ftype += typ.strip()

            t = ctypes_map.get(ftype)
            if t:
                ftype = f"{t.__module__}.{t.__name__}"

            flen = fmatch.group("length") or None
            c_fields.append((fname, ftype, flen))

        structs[name] = c_fields

    # Add a placeholder for signal definitions
    for msg_type in msg_types.keys():
        if msg_type.startswith("MT_"):
            structs.setdefault("MDF_" + msg_type[3:], None)

    return structs


def generate_constant(name: str, value: str):
    return f"{name} = {value}"


def generate_struct(name: str, fields):
    assert not name.startswith("MDF_")
    f = []
    fnum = len(fields)
    for i, (fname, ftyp, flen) in enumerate(fields, start=1):
        if flen and re.search(r"/|\+|\*|-", flen):
            flen = "int(" + flen + ")"
        nl = ",\n" if i < fnum else ""
        f.append(f"        (\"{fname}\", {ftyp}{' * ' + flen if flen else ''}){nl}")

    fstr = "".join(f)

    template = f"""
class {name}(ctypes.Structure):
    _fields_ = [
{fstr}
    ]

"""
    return template


def generate_msg_def(name: str, fields):
    assert name.startswith("MDF_")

    basename = name[4:]
    f = []
    fnum = len(fields)
    for i, (fname, ftyp, flen) in enumerate(fields, start=1):
        if flen and re.search(r"/|\+|\*|-", flen):
            flen = "int(" + flen + ")"
        nl = ",\n" if i < fnum else ""
        f.append(f"        (\"{fname}\", {ftyp}{' * ' + flen if flen else ''}){nl}")

    fstr = "".join(f)

    msg_id = "MT_" + basename
    template = f"""
@pyrtma.msg_def
class {name}(pyrtma.MessageData):
    _fields_ = [
{fstr}
    ]
    type_id = {msg_id}
    type_name = \"{basename}\"

"""
    return template


def generate_sig_def(name: str):
    assert name.startswith("MDF_")
    basename = name[4:]
    msg_id = "MT_" + basename
    template = f"""
# Signal Definition
@pyrtma.msg_def
class {name}(pyrtma.MessageData):
    _fields_ = []
    type_id = {msg_id}
    type_name = \"{basename}\"

"""
    return template


def print_content(content: str = "", end: str = "\n", out_file: Optional[IO] = None):
    """Print content to the console and optionally write to file"""
    if out_file and not out_file.closed:
        out_file.write(f"{content}{end}")
    else:
        print(content, end=end)


def parse_file(filename, seq: int = 1, out_filename: Optional[str] = None):
    """Parse a C header file for message definitions.
    Notes:
        * Does not follow other #includes
        * Parsing order: #defines, typedefs, typedef struct
    """

    with open(filename, "r") as f:
        text = f.read()

    out_file: Optional[IO] = None
    if out_filename:
        mode = "w" if seq == 1 else "a"
        out_file = open(out_filename, mode=mode)

    try:
        text = preprocess(text)
        defines = parse_defines(text)
        typedefs = parse_typedefs(text)
        structs = parse_structs(defines["MT"], text)

        if seq == 1:
            print_content("import ctypes", out_file=out_file)
            print_content("import pyrtma", out_file=out_file)
            print_content("from pyrtma.constants import *", out_file=out_file)
            print_content(out_file=out_file)

        print_content(f"# User Constants: {filename}", out_file=out_file)
        for name, value in defines["constants"].items():
            print_content(generate_constant(name, value), out_file=out_file)

        print_content(out_file=out_file)
        print_content(f"# User Message IDs: {filename}", out_file=out_file)
        for name, value in defines["MT"].items():
            print_content(generate_constant(name, value), out_file=out_file)

        print_content(out_file=out_file)
        print_content(f"# User Module IDs: {filename}", out_file=out_file)
        for name, value in defines["MID"].items():
            print_content(generate_constant(name, value), out_file=out_file)

        print_content(out_file=out_file)
        print_content(f"# User Type Definitions: {filename}", out_file=out_file)
        for name, value in typedefs.items():
            print_content(generate_constant(name, value), out_file=out_file)

        ctypes_map.update(typedefs)

        if not out_file:
            print_content()
        print_content(
            f"# User Message Definitions: {filename}", end="\n\n", out_file=out_file
        )

        for name, fields in structs.items():
            if name.startswith("MDF_"):
                if fields is not None:
                    print_content(
                        generate_msg_def(name, fields), end="", out_file=out_file
                    )
                else:
                    print_content(generate_sig_def(name), end="", out_file=out_file)
            else:
                print_content(generate_struct(name, fields), end="", out_file=out_file)

    finally:
        if out_file and not out_file.closed:
            out_file.close()


def python_v1_compile(include_files: List, out_filename: Optional[str]):
    for n, f in enumerate(include_files, start=1):
        parse_file(f, seq=n, out_filename=out_filename)



import pathlib
from pyrtma.parser import Parser
from ruamel.yaml import YAML
from typing import Dict, Any, Optional
from pyrtma.__version__ import __version__


class NullRepresenter:
    def __init__(self):
        self.count = 0

    def __call__(self, repr, data):
        return repr.represent_scalar("tag:yaml.org,2002:null", "null")


class YAMLCompiler:
    def __init__(self, parser: Parser, filename: str, debug: bool = False):
        self.filename = filename
        self.debug = debug
        self.parser = parser

    def generate(self, out_filepath: pathlib.Path):
        yaml = YAML()
        yaml.default_flow_style = False
        my_represent_none = NullRepresenter()
        yaml.representer.add_representer(type(None), my_represent_none)

        meta_dict: Dict[str, Any] = {
            "COMPILED_PYRTMA_VERSION": __version__,
            "AUTOGENERATED": True,
        }
        options_dict: Dict[str, Any] = {"IMPORT_COREDEFS": False}

        d: Dict[str, Optional[Dict[str, Any]]] = {
            "metadata": {},
            "compiler_options": {},
        }

        for k, v in self.parser.yaml_dict.items():
            if isinstance(v, dict) and len(v) == 0:
                d[k] = None
            else:
                d[k] = v
        for k, v in meta_dict.items():
            if d["metadata"] is None:
                d["metadata"] = {}
            if isinstance(d["metadata"], dict) and isinstance(v, dict) and len(v) == 0:
                d["metadata"][k] = None
            elif isinstance(d["metadata"], dict):
                d["metadata"][k] = v
        for k, v in options_dict.items():
            if d["compiler_options"] is None:
                d["compiler_options"] = {}
            if (
                isinstance(d["compiler_options"], dict)
                and isinstance(v, dict)
                and len(v) == 0
            ):
                d["compiler_options"][k] = None
            elif isinstance(d["compiler_options"], dict):
                d["compiler_options"][k] = v

        yaml.dump(d, out_filepath)



import unittest

# Import message defs to add to pyrtma.msg_defs map
from .test_msg_defs.test_defs import *


class TestEncoding(unittest.TestCase):
    """Test sending messages through MessageManager."""

    def test_validator_range(self):
        m = MDF_VALIDATOR_A()

        m.int8 = -128
        m.int8 = int(127)

        m.uint8 = 0
        m.uint8 = 255

        m.int16 = -32768
        m.int16 = 32767

        m.uint16 = 0
        m.uint16 = 65535

        m.int32 = -2_147_483_648
        m.int32 = 2_147_483_647

        m.uint32 = 0
        m.uint32 = 4_294_967_295

        m.int64 = -9_223_372_036_854_775_808
        m.int64 = 9_223_372_036_854_775_807

        m.uint64 = 0
        m.uint64 = 18_446_744_073_709_551_615

        m.float_ = 3.40282346638528859811704183484516925440e38
        m.float_ = -3.40282346638528859811704183484516925440e38

        m.double_ = 1.79769313486231570814527423731704356798070e308
        m.double_ = -1.79769313486231570814527423731704356798070e308

        m.byte_ = b"\x00"
        m.byte_ = b"\xff"

        m.char_ = chr(0)
        m.char_ = chr(127)

        m.int8_arr[0] = -128
        m.int8_arr[0] = 127

        m.uint8_arr[0] = 0
        m.uint8_arr[0] = 255

        m.int16_arr[0] = -32768
        m.int16_arr[0] = 32767

        m.uint16_arr[0] = 0
        m.uint16_arr[0] = 65535

        m.int32_arr[0] = -2_147_483_648
        m.int32_arr[0] = 2_147_483_647

        m.uint32_arr[0] = 0
        m.uint32_arr[0] = 4_294_967_295

        m.int64_arr[0] = -9_223_372_036_854_775_808
        m.int64_arr[0] = 9_223_372_036_854_775_807

        m.uint64_arr[0] = 0
        m.uint64_arr[0] = 18_446_744_073_709_551_615

        m.float_arr[0] = -3.40282346638528859811704183484516925440e38
        m.float_arr[0] = 3.40282346638528859811704183484516925440e38

        m.double_arr[0] = -1.79769313486231570814527423731704356798070e308
        m.double_arr[0] = 1.79769313486231570814527423731704356798070e308

        m.byte_arr[0] = b"\x00"
        m.byte_arr[0] = b"\xff"

    def test_validator_out_of_range(self):
        m = MDF_VALIDATOR_A()
        with self.assertRaises(ValueError):
            m.int8 = -129

        with self.assertRaises(ValueError):
            m.int8 = 128

        with self.assertRaises(ValueError):
            m.uint8 = -1

        with self.assertRaises(ValueError):
            m.uint8 = 256

        with self.assertRaises(ValueError):
            m.int16 = -32769

        with self.assertRaises(ValueError):
            m.int16 = 32768

        with self.assertRaises(ValueError):
            m.uint16 = -1

        with self.assertRaises(ValueError):
            m.uint16 = 65536

        with self.assertRaises(ValueError):
            m.int32 = -2_147_483_649

        with self.assertRaises(ValueError):
            m.int32 = 2_147_483_648

        with self.assertRaises(ValueError):
            m.uint32 = -1

        with self.assertRaises(ValueError):
            m.uint32 = 4_294_967_296

        with self.assertRaises(ValueError):
            m.int64 = -9_223_372_036_854_775_809

        with self.assertRaises(ValueError):
            m.int64 = 9_223_372_036_854_775_808

        with self.assertRaises(ValueError):
            m.uint64 = -1

        with self.assertRaises(ValueError):
            m.uint64 = 18_446_744_073_709_551_616

        with self.assertRaises(ValueError):
            m.float_ = -4.40282346638528859811704183484516925441e38

        with self.assertRaises(ValueError):
            m.float_ = 4.40282346638528859811704183484516925441e38

        with self.assertRaises(ValueError):
            m.double_ = 2.79769313486231570814527423731704356798071e308

        with self.assertRaises(ValueError):
            m.double_ = -2.79769313486231570814527423731704356798071e308

        with self.assertRaises(ValueError):
            m.byte_ = -1

        with self.assertRaises(ValueError):
            m.byte_ = 256

        # Arrays
        with self.assertRaises(ValueError):
            m.int8_arr[0] = -129

        with self.assertRaises(ValueError):
            m.int8_arr[0] = 128

        with self.assertRaises(ValueError):
            m.uint8_arr[0] = -1

        with self.assertRaises(ValueError):
            m.uint8_arr[0] = 256

        with self.assertRaises(ValueError):
            m.int16_arr[0] = -32769

        with self.assertRaises(ValueError):
            m.int16_arr[0] = 32768

        with self.assertRaises(ValueError):
            m.uint16_arr[0] = -1

        with self.assertRaises(ValueError):
            m.uint16_arr[0] = 65536

        with self.assertRaises(ValueError):
            m.int32_arr[0] = -2_147_483_649

        with self.assertRaises(ValueError):
            m.int32_arr[0] = 2_147_483_648

        with self.assertRaises(ValueError):
            m.uint32_arr[0] = -1

        with self.assertRaises(ValueError):
            m.uint32_arr[0] = 4_294_967_296

        with self.assertRaises(ValueError):
            m.int64_arr[0] = -9_223_372_036_854_775_809

        with self.assertRaises(ValueError):
            m.int64_arr[0] = 9_223_372_036_854_775_808

        with self.assertRaises(ValueError):
            m.uint64_arr[0] = -1

        with self.assertRaises(ValueError):
            m.uint64_arr[0] = 18_446_744_073_709_551_616

        with self.assertRaises(ValueError):
            m.float_arr[0] = -4.40282346638528859811704183484516925441e38

        with self.assertRaises(ValueError):
            m.float_arr[0] = 4.40282346638528859811704183484516925441e38

        with self.assertRaises(ValueError):
            m.double_arr[0] = 2.79769313486231570814527423731704356798071e308

        with self.assertRaises(ValueError):
            m.double_arr[0] = -2.79769313486231570814527423731704356798071e308

        with self.assertRaises(ValueError):
            m.byte_arr[0] = -1

        with self.assertRaises(ValueError):
            m.byte_arr[0] = 256

    def test_array_len(self):
        m = MDF_VALIDATOR_A()

        with self.assertRaises(ValueError):
            m.int8_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int8_arr[:] = (ctypes.c_int8 * 5)()

        with self.assertRaises(ValueError):
            m.uint8_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint8_arr[:] = (ctypes.c_uint8 * 5)()

        with self.assertRaises(ValueError):
            m.int16_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int16_arr[:] = (ctypes.c_int16 * 5)()

        with self.assertRaises(ValueError):
            m.uint16_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint16_arr[:] = (ctypes.c_uint16 * 5)()

        with self.assertRaises(ValueError):
            m.int32_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int32_arr[:] = (ctypes.c_int32 * 5)()

        with self.assertRaises(ValueError):
            m.uint32_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint32_arr[:] = (ctypes.c_uint32 * 5)()

        with self.assertRaises(ValueError):
            m.int64_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int64_arr[:] = (ctypes.c_int64 * 5)()

        with self.assertRaises(ValueError):
            m.uint64_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint64_arr[:] = (ctypes.c_uint64 * 5)()

        with self.assertRaises(ValueError):
            m.float_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.float_arr[:] = (ctypes.c_float * 5)()

        with self.assertRaises(ValueError):
            m.double_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.double_arr[:] = (ctypes.c_double * 5)()

        with self.assertRaises(ValueError):
            m.byte_arr[:] = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.byte_arr[:] = (ctypes.c_ubyte * 5)()

        # test again without [:]
        with self.assertRaises(ValueError):
            m.int8_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int8_arr = (ctypes.c_int8 * 5)()

        with self.assertRaises(ValueError):
            m.uint8_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint8_arr = (ctypes.c_uint8 * 5)()

        with self.assertRaises(ValueError):
            m.int16_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int16_arr = (ctypes.c_int16 * 5)()

        with self.assertRaises(ValueError):
            m.uint16_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint16_arr = (ctypes.c_uint16 * 5)()

        with self.assertRaises(ValueError):
            m.int32_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int32_arr = (ctypes.c_int32 * 5)()

        with self.assertRaises(ValueError):
            m.uint32_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint32_arr = (ctypes.c_uint32 * 5)()

        with self.assertRaises(ValueError):
            m.int64_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.int64_arr = (ctypes.c_int64 * 5)()

        with self.assertRaises(ValueError):
            m.uint64_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.uint64_arr = (ctypes.c_uint64 * 5)()

        with self.assertRaises(ValueError):
            m.float_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.float_arr = (ctypes.c_float * 5)()

        with self.assertRaises(ValueError):
            m.double_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.double_arr = (ctypes.c_double * 5)()

        with self.assertRaises(ValueError):
            m.byte_arr = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            m.byte_arr = (ctypes.c_ubyte * 5)()

    def test_get_types(self):
        m = MDF_VALIDATOR_A()
        self.assertIsInstance(m.int8, int)
        self.assertIsInstance(m.int16, int)
        self.assertIsInstance(m.int32, int)
        self.assertIsInstance(m.int64, int)
        self.assertIsInstance(m.uint8, int)
        self.assertIsInstance(m.uint16, int)
        self.assertIsInstance(m.uint32, int)
        self.assertIsInstance(m.uint64, int)
        self.assertIsInstance(m.float_, float)
        self.assertIsInstance(m.double_, float)
        self.assertIsInstance(m.byte_, int)
        self.assertIsInstance(m.char_, str)
        self.assertIsInstance(m.string, str)
        self.assertIsInstance(m.struct_, VALIDATOR_STRUCT)

        self.assertIsInstance(m.int8_arr, IntArray)
        self.assertIsInstance(m.int16_arr, IntArray)
        self.assertIsInstance(m.int32_arr, IntArray)
        self.assertIsInstance(m.int64_arr, IntArray)
        self.assertIsInstance(m.uint8_arr, IntArray)
        self.assertIsInstance(m.uint16_arr, IntArray)
        self.assertIsInstance(m.uint32_arr, IntArray)
        self.assertIsInstance(m.uint64_arr, IntArray)
        self.assertIsInstance(m.float_arr, FloatArray)
        self.assertIsInstance(m.double_arr, FloatArray)
        self.assertIsInstance(m.byte_arr, ByteArray)
        self.assertIsInstance(m.struct_arr, StructArray)

        # Single value
        self.assertIsInstance(m.int8_arr[0], int)
        self.assertIsInstance(m.int16_arr[0], int)
        self.assertIsInstance(m.int32_arr[0], int)
        self.assertIsInstance(m.int64_arr[0], int)
        self.assertIsInstance(m.uint8_arr[0], int)
        self.assertIsInstance(m.uint16_arr[0], int)
        self.assertIsInstance(m.uint32_arr[0], int)
        self.assertIsInstance(m.uint64_arr[0], int)

        # Slice
        self.assertIsInstance(m.int8_arr[0:2], list)
        self.assertIsInstance(m.int8_arr[0:2][0], int)

        self.assertIsInstance(m.int16_arr[0:2], list)
        self.assertIsInstance(m.int16_arr[0:2][0], int)

        self.assertIsInstance(m.int32_arr[0:2], list)
        self.assertIsInstance(m.int32_arr[0:2][0], int)

        self.assertIsInstance(m.int64_arr[0:2], list)
        self.assertIsInstance(m.int64_arr[0:2][0], int)

        self.assertIsInstance(m.uint8_arr[0:2], list)
        self.assertIsInstance(m.uint8_arr[0:2][0], int)

        self.assertIsInstance(m.uint16_arr[0:2], list)
        self.assertIsInstance(m.uint16_arr[0:2][0], int)

        self.assertIsInstance(m.uint32_arr[0:2], list)
        self.assertIsInstance(m.uint32_arr[0:2][0], int)

        self.assertIsInstance(m.uint64_arr[0:2], list)
        self.assertIsInstance(m.uint64_arr[0:2][0], int)

        self.assertIsInstance(m.byte_arr[0:2], bytearray)
        self.assertIsInstance(m.byte_arr[0:2][0], int)

        self.assertIsInstance(m.struct_arr[0:2], list)
        self.assertIsInstance(m.struct_arr[0:2][0], VALIDATOR_STRUCT)

    def test_invalid_set_types(self):
        A = MDF_VALIDATOR_A()
        B = MDF_VALIDATOR_B()

        with self.assertRaises(TypeError):
            A.char_ = 1

        with self.assertRaises(TypeError):
            A.char_ = ctypes.c_int8(1)

        with self.assertRaises(TypeError):
            A.char_ = ctypes.c_byte(1)

        with self.assertRaises(TypeError):
            A.char_ = ctypes.c_ubyte(1)

        with self.assertRaises(TypeError):
            A.int8 = "a"

        with self.assertRaises(TypeError):
            A.int8 = ctypes.c_int16(1)

        with self.assertRaises(TypeError):
            A.int8 = ctypes.c_uint8(1)

        with self.assertRaises(TypeError):
            A.int8 = 1.4

        with self.assertRaises(TypeError):
            A.int16 = [1, 2, 3]

        with self.assertRaises(TypeError):
            A.int32 = b"1"

        with self.assertRaises(TypeError):
            A.int64 = dict()

        with self.assertRaises(TypeError):
            A.uint8 = "abc"

        with self.assertRaises(TypeError):
            A.uint16 = object()

        with self.assertRaises(TypeError):
            A.uint32 = "a"

        with self.assertRaises(TypeError):
            A.uint64 = "a"

        with self.assertRaises(TypeError):
            A.float_ = "a"

        with self.assertRaises(TypeError):
            A.double_ = b"1"

        with self.assertRaises(TypeError):
            A.byte_ = "a"

        with self.assertRaises(TypeError):
            A.byte_ = ctypes.c_char(1)

        with self.assertRaises(TypeError):
            A.string = 1

        # Int8
        with self.assertRaises(ValueError):
            A.int8_arr = B.int8_arr

        with self.assertRaises(TypeError):
            A.int8_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.int8_arr = IntArray(Int8, 4)

        # Int16
        with self.assertRaises(ValueError):
            A.int16_arr = B.int16_arr

        with self.assertRaises(TypeError):
            A.int16_arr = IntArray(Int8, 4)

        with self.assertRaises(ValueError):
            A.int16_arr = IntArray(Int16, 4)

        # Int32
        with self.assertRaises(ValueError):
            A.int32_arr = B.int32_arr

        with self.assertRaises(TypeError):
            A.int32_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.int32_arr = IntArray(Int32, 4)

        # Int64
        with self.assertRaises(ValueError):
            A.int64_arr = B.int64_arr

        with self.assertRaises(TypeError):
            A.int64_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.int64_arr = IntArray(Int64, 4)

        # Uint8
        with self.assertRaises(ValueError):
            A.uint8_arr = B.uint8_arr

        with self.assertRaises(TypeError):
            A.uint8_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.uint8_arr = IntArray(Uint8, 4)

        # Uint16
        with self.assertRaises(ValueError):
            A.uint16_arr = B.uint16_arr

        with self.assertRaises(TypeError):
            A.uint16_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.uint16_arr = IntArray(Uint16, 4)

        # Uint32
        with self.assertRaises(ValueError):
            A.uint32_arr = B.uint32_arr

        with self.assertRaises(TypeError):
            A.uint32_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.uint32_arr = IntArray(Uint32, 4)

        # Uint64
        with self.assertRaises(ValueError):
            A.uint64_arr = B.uint64_arr

        with self.assertRaises(TypeError):
            A.uint64_arr = IntArray(Int16, 4)

        with self.assertRaises(ValueError):
            A.uint64_arr = IntArray(Uint64, 4)

        # Float
        with self.assertRaises(ValueError):
            A.float_arr = B.float_arr

        with self.assertRaises(TypeError):
            A.float_arr = FloatArray(Double, 4)

        with self.assertRaises(ValueError):
            A.float_arr = FloatArray(Float, 4)

        # Double
        with self.assertRaises(ValueError):
            A.double_arr = B.double_arr

        with self.assertRaises(TypeError):
            A.double_arr = FloatArray(Float, 4)

        with self.assertRaises(ValueError):
            A.double_arr = FloatArray(Double, 4)

        # Byte
        with self.assertRaises(ValueError):
            A.byte_arr = B.byte_arr

        with self.assertRaises(ValueError):
            A.byte_arr = ByteArray(4)

        # Struct
        with self.assertRaises(TypeError):
            A.struct_arr = [1, 2, 3, 4]

        with self.assertRaises(ValueError):
            A.struct_arr = B.struct_arr

        with self.assertRaises(TypeError):
            A.struct_arr = StructArray(MDF_VALIDATOR_B, 4)

        with self.assertRaises(ValueError):
            A.struct_arr = StructArray(VALIDATOR_STRUCT, 4)

    def test_valid_set_types(self):
        A = MDF_VALIDATOR_A()
        A.char_ = "a"
        A.char_ = ctypes.c_char(98)
        A.int8 = 1
        A.int8 = ctypes.c_int8(2)
        A.int16 = 1
        A.int16 = ctypes.c_int16(2)
        A.int32 = 1
        A.int32 = ctypes.c_int32(2)
        A.int64 = 1
        A.int64 = ctypes.c_int64(2)
        A.uint8 = 1
        A.uint8 = ctypes.c_uint8(2)
        A.uint16 = 1
        A.uint16 = ctypes.c_uint16(2)
        A.uint32 = 1
        A.uint32 = ctypes.c_uint32(2)
        A.uint64 = 1
        A.uint64 = ctypes.c_uint64(2)

        A.float_ = 1.0
        A.double_ = 1.0

        A.float_ = 1
        A.double_ = 1

        A.float_ = ctypes.c_float(1.1)
        A.double_ = ctypes.c_double(1.8)

        A.byte_ = 1
        A.byte_ = b"A"
        A.byte_ = bytes(1)
        A.byte_ = bytearray(1)
        A.byte_ = ctypes.c_ubyte(1)

        A.string = "abc"

        A.int8_arr[0] = 1
        A.int16_arr[0] = 1
        A.int32_arr[0] = 1
        A.int64_arr[0] = 1
        A.uint8_arr[0] = 1
        A.uint16_arr[0] = 1
        A.uint32_arr[0] = 1
        A.uint64_arr[0] = 1
        A.float_arr[0] = 1
        A.double_arr[0] = 1

        A.float_arr[0] = 1.0
        A.double_arr[0] = 1.0

        A.byte_arr[0] = 1
        A.byte_arr[0] = b"A"
        A.byte_arr[0] = bytes(1)
        A.byte_arr[0] = bytearray(1)

        A.struct_arr[0] = VALIDATOR_STRUCT()

        A.int8_arr[:] = [1, 2, 3, 4]
        A.int8_arr[:] = (ctypes.c_int8 * 4)()
        A.int16_arr[:] = [1, 2, 3, 4]
        A.int32_arr[:] = [1, 2, 3, 4]
        A.int64_arr[:] = [1, 2, 3, 4]
        A.uint8_arr[:] = [1, 2, 3, 4]
        A.uint16_arr[:] = [1, 2, 3, 4]
        A.uint32_arr[:] = [1, 2, 3, 4]
        A.uint64_arr[:] = [1, 2, 3, 4]
        A.float_arr[:] = [1.0, 2.0, 3.0, 4.0]
        A.double_arr[:] = [1.0, 2.0, 3.0, 4.0]

        A.float_arr[:] = [1, 2, 3, 4]
        A.double_arr[:] = [1, 2, 3, 4]

        A.byte_arr[:] = [1, 2, 3, 4]
        A.byte_arr[:] = b"ABCD"
        A.byte_arr[:] = bytes(4)
        A.byte_arr[:] = bytearray(4)

        A.struct_arr[:] = [
            VALIDATOR_STRUCT(),
            VALIDATOR_STRUCT(),
            VALIDATOR_STRUCT(),
            VALIDATOR_STRUCT(),
        ]

        A.struct_arr[:] = (VALIDATOR_STRUCT * 4)()

        # again without [:]
        A.int8_arr = [1, 2, 3, 4]
        A.int8_arr = (ctypes.c_int8 * 4)()
        A.int16_arr = [1, 2, 3, 4]
        A.int32_arr = [1, 2, 3, 4]
        A.int64_arr = [1, 2, 3, 4]
        A.uint8_arr = [1, 2, 3, 4]
        A.uint16_arr = [1, 2, 3, 4]
        A.uint32_arr = [1, 2, 3, 4]
        A.uint64_arr = [1, 2, 3, 4]
        A.float_arr = [1.0, 2.0, 3.0, 4.0]
        A.float_arr = (ctypes.c_float * 4)()
        A.double_arr = [1.0, 2.0, 3.0, 4.0]

        A.float_arr = [1, 2, 3, 4]
        A.double_arr = [1, 2, 3, 4]

        A.byte_arr = [1, 2, 3, 4]
        A.byte_arr = b"ABCD"
        A.byte_arr = bytes(4)
        A.byte_arr = bytearray(4)

        A.struct_arr = [
            VALIDATOR_STRUCT(),
            VALIDATOR_STRUCT(),
            VALIDATOR_STRUCT(),
            VALIDATOR_STRUCT(),
        ]

        A.struct_arr = (VALIDATOR_STRUCT * 4)()

    def test_refs(self):
        A = MDF_VALIDATOR_A()
        B = MDF_VALIDATOR_A()

        A.char_ = "a"
        A.int8 = 1
        A.int16 = 1
        A.int32 = 1
        A.int64 = 1
        A.uint8 = 1
        A.uint16 = 1
        A.uint32 = 1
        A.uint64 = 1
        A.float_ = 1.0
        A.double_ = 1.0
        A.byte_ = 1
        A.string = "abc"

        A.int8_arr[0] = 1
        A.int16_arr[0] = 1
        A.int32_arr[0] = 1
        A.int64_arr[0] = 1
        A.uint8_arr[0] = 1
        A.uint16_arr[0] = 1
        A.uint32_arr[0] = 1
        A.uint64_arr[0] = 1
        A.float_arr[0] = 1
        A.double_arr[0] = 1
        A.byte_arr[0] = 1
        A.struct_arr[0].char_ = "a"

        self.assertFalse(A.char_ == B.char_)
        self.assertFalse(A.int8 == B.int8)
        self.assertFalse(A.int16 == B.int16)
        self.assertFalse(A.int32 == B.int32)
        self.assertFalse(A.int64 == B.int64)
        self.assertFalse(A.uint8 == B.uint8)
        self.assertFalse(A.uint16 == B.uint16)
        self.assertFalse(A.uint32 == B.uint32)
        self.assertFalse(A.uint64 == B.uint64)
        self.assertFalse(A.float_ == B.float_)
        self.assertFalse(A.double_ == B.double_)
        self.assertFalse(A.byte_ == B.byte_)
        self.assertFalse(A.string == B.string)
        self.assertFalse(A.int8_arr == B.int8_arr)
        self.assertFalse(A.int16_arr == B.int16_arr)
        self.assertFalse(A.int32_arr == B.int32_arr)
        self.assertFalse(A.int64_arr == B.int64_arr)
        self.assertFalse(A.uint8_arr == B.uint8_arr)
        self.assertFalse(A.uint16_arr == B.uint16_arr)
        self.assertFalse(A.uint32_arr == B.uint32_arr)
        self.assertFalse(A.uint64_arr == B.uint64_arr)
        self.assertFalse(A.float_arr == B.float_arr)
        self.assertFalse(A.double_arr == B.double_arr)
        self.assertFalse(A.byte_arr == B.byte_arr)
        self.assertFalse(A.struct_arr == B.struct_arr)

        self.assertFalse(A.int8_arr is B.int8_arr)
        self.assertFalse(A.int16_arr is B.int16_arr)
        self.assertFalse(A.int32_arr is B.int32_arr)
        self.assertFalse(A.int64_arr is B.int64_arr)
        self.assertFalse(A.uint8_arr is B.uint8_arr)
        self.assertFalse(A.uint16_arr is B.uint16_arr)
        self.assertFalse(A.uint32_arr is B.uint32_arr)
        self.assertFalse(A.uint64_arr is B.uint64_arr)
        self.assertFalse(A.float_arr is B.float_arr)
        self.assertFalse(A.double_arr is B.double_arr)
        self.assertFalse(A.byte_arr is B.byte_arr)
        self.assertFalse(A.struct_arr is B.struct_arr)



import datetime
import json
import re
import os

from .exceptions import MissingMetadata, InvalidMetadata
from typing import Union


class LoggingMetadata:
    def __init__(self) -> None:
        self._metadata: dict[str, Union[str, float, int]] = {}

    def update(self, json_str: str):
        try:
            new_data = json.loads(json_str)
        except json.JSONDecodeError:
            raise InvalidMetadata("The metadata provided was not formatted as json.")

        if not all(isinstance(v, (str, float, int)) for v in new_data.values()):
            raise InvalidMetadata("Metadata can only contain str, float, or ints.")

        self._metadata.clear()
        self._metadata.update(**new_data)
        self._update_builtins()

    def to_json(self):
        return json.dumps(self._metadata, separators=(",", ":"))

    def clear(self):
        self._metadata.clear()

    def _update_builtins(self):
        now = datetime.datetime.now()

        # Builtin Time metadata
        self._metadata.update(
            date=now.strftime("%Y_%m_%d"),
            isodate=now.strftime("%Y-%m-%d"),
            timestamp=now.strftime("%H%M%S"),
            timestamp_ms=now.strftime("%H%M%S%f"),
        )

    def expand_format(self, s: str) -> str:
        self._update_builtins()

        def repl(m: re.Match) -> str:
            d = m.groupdict()
            key = d["key"]
            fmt = d["fmt"] or ""

            if d["scope"] == "env":
                try:
                    value = os.environ[key]
                except KeyError:
                    raise MissingMetadata(
                        f"No value found in env for {key} in metadata format string: {s}"
                    )
            else:
                try:
                    value = self._metadata[key]
                except KeyError:
                    raise MissingMetadata(
                        f"No value found in metadata for {key} in metadata format string: {s}"
                    )

            return value.__format__(fmt)

        pattern = r"\$\(((?P<scope>env):)?(?P<key>\w+)(:(?P<fmt>\w*))?\)"
        s = re.sub(pattern, repl, s)

        if s.find("$") != -1:
            raise MissingMetadata(
                f"Failed to fully expand format string with metadata: {s}"
            )

        return s



import ctypes
import json

from typing import TypeVar, Any, Type, Dict, List
from .utils.print import print_ctype_array, hexdump
from .utils.random_fields import _random_struct
from .exceptions import JSONDecodingError

MB = TypeVar("MB", bound="MessageBase")


# MessageBase Metaclass - (for runtime ctypes field generation)
CStructType: type = type(
    ctypes.Structure
)  # technically incorrect typehint, but makes mypy happy


class MessageMeta(CStructType):
    """MessageMeta metaclass

    Responsible for generating ctypes fields from descriptor attributes prior to class creation
    """

    def __new__(cls, name, bases, namespace):
        fields = []
        for key in namespace.keys():
            if hasattr(namespace[key], "_ctype"):
                fname = "_" + key
                ftype = namespace[key]._ctype
                fields.append((fname, ftype))
            elif key == "_fields_":
                fields.extend(namespace["_fields_"])
        if fields:
            namespace["_fields_"] = fields
        elif "_fields_" not in namespace:
            # else important to not overwrite v1 message def fields
            namespace["_fields_"] = []

        return super().__new__(cls, name, bases, namespace)


# "abstract" base class for MessageHeader and MessageData
class MessageBase(ctypes.Structure, metaclass=MessageMeta):
    """MessageBase base class

    This class should be treated as if abstract and not instantiated directly.
    """

    @property
    def size(self) -> int:
        return ctypes.sizeof(self)

    def pretty_print(self, add_tabs=0) -> str:
        """Generate formatted message structure string for pretty printing

        Args:
            add_tabs (int, optional): Indentation level, used for recursively calling. Defaults to 0.

        Returns:
            str: Formatted string
        """
        max_len = 20
        pstr = "\t" * add_tabs + f"{type(self).__name__}:"
        for field_name, field_type, *_ in self._fields_:
            if field_name[0] == "_":
                field_name = field_name[1:]
            val = getattr(self, field_name)
            class_name = field_type.__name__
            if type(val) is str:
                sval = f'"{val}"'
            elif hasattr(val, "pretty_print"):
                sval = val.pretty_print(add_tabs + 1)
            # expand arrays
            elif hasattr(val, "__len__"):
                if hasattr(val, "_type_"):
                    class_name = val._type_.__name__
                if hasattr(val, "pretty_print"):
                    sval = val.pretty_print(add_tabs + 1)
                elif (
                    type(val) not in (bytes, bytearray)
                    and type(val[0]) is not bytearray
                ):
                    sval = print_ctype_array(val)
                elif len(val) > max_len:
                    sval = f"{val[:max_len]}..."
                else:
                    sval = str(val)
            else:
                sval = str(val)
            if hasattr(val, "pretty_print"):
                pstr += (
                    f"\n"
                    + "\t" * (add_tabs + 1)
                    + f"{field_name} = ({class_name})\n{sval}"
                )
            else:
                pstr += (
                    f"\n"
                    + "\t" * (add_tabs + 1)
                    + f"{field_name} = ({class_name}){sval}"
                )
        return pstr

    def hexdump(self, length=16, sep=" "):
        """hexdump of message

        Args:
            length (int, optional): Row length. Defaults to 16.
            sep (str, optional): Separator for non-printable ascii chars. Defaults to " ".
        """
        hexdump(bytes(self), length, sep)

    def to_dict(self) -> Dict[str, Any]:
        """Convert message to dictionary

        Returns:
            Dict[str, Any]: Message dictionary
        """
        return _to_dict(self)

    def to_json(self, minify: bool = False, **kwargs) -> str:
        """Convert message to json string

        Args:
            minify (bool, optional): Flag to minify (compact format). Defaults to False.
            kwargs for json.dumps

        Returns:
            str: json string
        """
        if minify:
            return json.dumps(
                self, cls=RTMAJSONEncoder, separators=(",", ":"), **kwargs
            )
        else:
            return json.dumps(self, cls=RTMAJSONEncoder, indent=2, **kwargs)

    def get_field_raw(self, name: str) -> bytes:
        """return copy of raw bytes for ctypes field

        Args:
            name (str): Message fieldname

        Raises:
            KeyError: Invalid fieldname

        Returns:
            bytes: Copy of message field data bytes
        """
        meta = getattr(type(self), f"_{name}")
        if name not in [x[0] for x in self._fields_]:
            raise KeyError(f"{name} is not a field of {self.__name__}")
        else:
            offset = meta.offset
            sz = meta.size

        return bytes(memoryview(self).cast("c")[offset : offset + sz])

    @classmethod
    def from_dict(cls: Type[MB], data: Dict[str, Any]) -> MB:
        """Generate message instance from dictionary

        Args:
            data (Dict[str, Any]): Message dictionary
        Raises:
            JSONDecodingError: Unable to decode dictionary
        """
        obj = cls()
        try:
            _from_dict(obj, data)
            return obj
        except Exception as e:
            raise JSONDecodingError(
                f"Unable to decode {type(obj).__name__} from {data}"
            )

    @classmethod
    def from_json(cls: Type[MB], s: str) -> MB:
        """Generate message instance from JSON string

        Args:
            s (str): Message JSON string
        """
        obj = cls.from_dict(json.loads(s))
        return obj

    @classmethod
    def from_random(cls: Type[MB]) -> MB:
        """Generate message instance with random values"""
        obj = _random_struct(cls())
        return obj

    @classmethod
    def copy(cls: Type[MB], m: MB) -> MB:
        """Generate a copy of a message structure

        Args:
            m: Message structure to copy
        """
        return cls.from_buffer_copy(m)

    def __str__(self) -> str:
        return self.pretty_print()

    def __eq__(self, other: Any) -> bool:
        if type(self) != type(other):
            return False

        return bytes(self) == bytes(other)

    def __getattribute__(self, name: str) -> Any:
        # for backwards compatibility with messages without descriptors
        # can be deprecated
        value = super().__getattribute__(name)
        # need to convert chars from bytes to str
        if type(value) is bytes:
            fnames, ftypes = zip(*self._fields_)
            if name in fnames and name[0] != "_":  # v1 fields
                i = fnames.index(name)
                ftype = ftypes[i]
                if ftype is ctypes.c_char or (
                    issubclass(ftype, ctypes.Array) and ftype._type_ is ctypes.c_char
                ):
                    return value.decode()
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        # for backwards compatibility with messages without descriptors
        # can be deprecated
        if type(value) is str:
            fnames, ftypes = zip(*self._fields_)
            if name in fnames and name[0] != "_":  # v1 fields
                i = fnames.index(name)
                ftype = ftypes[i]
                if ftype is ctypes.c_char or (
                    issubclass(ftype, ctypes.Array) and ftype._type_ is ctypes.c_char
                ):
                    value = value.encode()
        super().__setattr__(name, value)


class RTMAJSONEncoder(json.JSONEncoder):
    """JSONEncoder object used to convert MessageData to json

    Example:
        data = json.dumps(msg, cls=pyrtma.encoding.RTMAJSONEncoder)
    """

    def default(self, o: Any) -> Any:
        """Default method to return serializable objects

        Args:
            o (Any): data to serialize

        Returns:
            Any: Serializable data
        """
        if isinstance(o, MessageBase) or (hasattr(o, "to_dict")):
            return o.to_dict()

        if isinstance(o, ctypes.Array):
            return list(o)

        if isinstance(o, (bytes, bytearray)):
            return [int(x) for x in o]

        return super().default(o)


def _from_dict(obj: MessageBase, data: Dict[str, Any]):
    """Helper function to set message fields from dictionary values

    Args:
        obj (MessageBase): Message object
        data (Dict[str, Any]): Message data dictionary
    """
    for _name, ftype, *_ in obj._fields_:
        name = _name[1:] if _name[0] == "_" else _name
        if issubclass(ftype, MessageBase):
            _from_dict(getattr(obj, name), data[name])
        elif issubclass(ftype, ctypes.Array):
            if issubclass(ftype._type_, MessageBase):  # type: ignore
                for i, elem in enumerate(getattr(obj, name)):
                    _from_dict(elem, data[name][i])
            elif ftype._type_ is ctypes.c_char:
                # list of characters is equivalent to str
                if type(data[name]) is list and all(
                    [(type(x) is str and len(x) <= 1) for x in data[name]]
                ):
                    data[name] = "".join(data[name])

                setattr(obj, name, data[name])
            else:
                getattr(obj, name)[:] = data[name]
        else:
            setattr(obj, name, data[name])


def _expand_carray(array: ctypes.Array) -> List:
    expanded = array[:]
    for i, a in enumerate(expanded):
        if isinstance(a, ctypes.Array):
            expanded[i] = _expand_carray(expanded[i])
    return expanded


def _to_dict(obj: MessageBase) -> Dict[str, Any]:
    """Helper function to create dictionary from message fields

    Args:
        obj (MessageBase): Message object

    Returns:
        Dict[str, Any]: Dictionary
    """
    data: Dict[str, Any] = {}
    for _name, ftype, *_ in obj._fields_:
        name = _name[1:] if _name[0] == "_" else _name
        if issubclass(ftype, MessageBase):
            data[name] = _to_dict(getattr(obj, name))
        elif issubclass(ftype, ctypes.Array):
            # Struct
            if issubclass(ftype._type_, MessageBase):  # type: ignore
                data[name] = []
                for elem in getattr(obj, name):
                    data[name].append(_to_dict(elem))

            # String
            elif ftype._type_ is ctypes.c_char:
                data[name] = getattr(obj, name)

            # Int8 Array
            elif ftype._type_ is ctypes.c_byte:
                data[name] = getattr(obj, name)[:].copy()

            # Uint8 or Byte Array
            elif ftype._type_ is ctypes.c_ubyte:
                if type(getattr(obj, name)).__name__ == "ByteArray":
                    # ByteArray
                    data[name] = bytes(getattr(obj, name)[:])
                else:
                    # Uint8Array
                    data[name] = getattr(obj, name)[:].copy()

            # Any other Array Type
            else:
                data[name] = _expand_carray(getattr(obj, name))

        else:
            data[name] = getattr(obj, name)

    return data
